// Copyright (C) 2010 - 2016 by Pedro Mendes, Virginia Tech Intellectual
// Properties, Inc., University of Heidelberg, and The University
// of Manchester.
// All rights reserved.

// Copyright (C) 2008 - 2009 by Pedro Mendes, Virginia Tech Intellectual
// Properties, Inc., EML Research, gGmbH, University of Heidelberg,
// and The University of Manchester.
// All rights reserved.

// Copyright (C) 2004 - 2007 by Pedro Mendes, Virginia Tech Intellectual
// Properties, Inc. and EML Research, gGmbH.
// All rights reserved.

#ifndef OUTPUT_HANDLER
#define OUTPUT_HANDLER

#include <vector>
#include <set>

#include "report/CCopasiObject.h"

#include "math/CMathUpdateSequence.h"

class CMathContainer;
class CCopasiContainer;
class CCopasiTask;
class Refresh;
class CCopasiDataModel;

/**
 *  This is just the interface that is used to all output provided by COPASI.
 */
class COutputInterface
{
public:
  /**
   * The output activity, indicating the status of the current operations
   * performed by a task.
   */
  enum Activity
  {
    BEFORE = 0x01,
    DURING = 0x02,
    AFTER = 0x04
  };

  /**
   * Default constructor
   */
  COutputInterface():
    mObjects()
  {};

  /**
   * Copy constructor
   * @param const COutputInterface & src
   */
  COutputInterface(const COutputInterface & src):
    mObjects(src.mObjects)
  {};

  /**
   * Destructor
   */
  virtual ~COutputInterface() {};

  /**
   * compile the object list from name vector
   * @param CObjectInterface::ContainerList listOfContainer
   * @return bool success
   */
  virtual bool compile(CObjectInterface::ContainerList /* listOfContainer */)
  {return false;};

  /**
   * Perform an output event for the current activity
   * @param const Activity & activity
   */
  virtual void output(const Activity & /* activity */) {};

  /**
   * Introduce an additional separator into the output
   * @param const Activity & activity
   */
  virtual void separate(const Activity & /* activity */) {};

  /**
   * Finish the output
   */
  virtual void finish() {};

  /**
   * Close the stream if applicable
   */
  virtual void close() {};

  /**
   * Retrieve the list of objects handled by the interface
   * @return const CObjectInterface::ObjectSet & objects
   */
  virtual const CObjectInterface::ObjectSet & getObjects() const
  {return mObjects;}

  // Attributes
protected:
  /**
   * All the objects which are output.
   */
  CObjectInterface::ObjectSet mObjects;
};

/**
 *  This is the lass which drives all output of COPASI.
 */
class COutputHandler: public COutputInterface
{
  // Operations

public:
  /**
   * Default Constructor
   */
  COutputHandler();

  /**
   * Copy constructor
   * @param const COutputInterface & src
   */
  COutputHandler(const COutputHandler & src);

  /**
   * Destructor
   */
  virtual ~COutputHandler();

  /**
   * compile the object list from name vector
   * @param CObjectInterface::ContainerList listOfContainer
   * @param  const CCopasiDataModel* pDataModel
   * @return bool success
   */
  virtual bool compile(CObjectInterface::ContainerList listOfContainer);

  /**
   * Perform an output event for the current activity
   * @param const Activity & activity
   */
  virtual void output(const Activity & activity);

  /**
   * Introduce an additional separator into the output
   * @param const Activity & activity
   */
  virtual void separate(const Activity & activity);

  /**
   * Finish the output
   */
  virtual void finish();

  /**
   * Add an interface
   * @param COutputInterface * pInterface;
   */
  virtual void addInterface(COutputInterface * pInterface);

  /**
   * Remove an interface
   * @param COutputInterface * pInterface;
   */
  virtual void removeInterface(COutputInterface * pInterface);

  /**
   * Set whether the handler is the master handler
   * @param COutputHandler * pMaster
   */
  void setMaster(COutputHandler * pMaster);

  /**
   * Check whether the handler is a master
   * @return bool isMaster
   */
  bool isMaster() const;

  std::set<COutputInterface *> getInterfaces() const;

protected:
  /**
   * Refresh all objects
   */
  void applyUpdateSequence();

  /**
   * Compile the object refresh list
   * @param const CObjectInterface::ContainerList & listOfContainer
   * @return bool success
   */
  bool compileUpdateSequence(const CObjectInterface::ContainerList & listOfContainer);

  // Attributes
protected:
  /**
   * A list of all active output interfaces.
   */
  std::set< COutputInterface * > mInterfaces;

  /**
   * Points to the master handler. The master handler is responsible for the
   * and object updates and all the output.
   */
  COutputHandler * mpMaster;

  /**
   * An ordered list of refresh methods needed by the master
   */
  CObjectInterface::UpdateSequence mUpdateSequence;

  /**
   * A pointer to the math container
   */
  const CMathContainer * mpContainer;
};
#endif
